//================================================================================================
//                                  cIotModuleSte body
//================================================================================================

void cIotModuleSte::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  
  //  pointer to App object used to address Generic* functions
  this->pIotModuleApp = pIotModuleApp;
  
  //  init average loop data
  OldUS = esp_timer_get_time();
  AverageLoopTimeMS = cIotModuleCom::TIMER_MS;
  
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    Data.Params[Channel].Attached = false;
  }
}

//  called from cIotModuleApp
bool cIotModuleSte::ConfigSpeedAcc(uint8_t Channel, uint32_t SpeedHz, uint16_t AccHz_mSec)
{
  bool RetVal = true;
  if (Channel != 0xff)
  {
    if (Data.Params[Channel % MAX_CHANNELS].Attached && SpeedHz <= MAX_SPEED && SpeedHz >= START_SPEED)
    {
      Data.Params[Channel].SpeedHz = SpeedHz;
      Data.Params[Channel].AccHz_mSec = AccHz_mSec;
    }
    else RetVal = false;
  }
  return RetVal;
}

//  called from cIotModuleApp
uint32_t IRAM_ATTR cIotModuleSte::GetDistanceFromDestination(const uint8_t Channel)
{
  uint32_t Value = 0;
  if (Channel != 0xff)
  {
    if (Data.Params[Channel % MAX_CHANNELS].Attached)
    {
      Value = AddOffset(Data.Params[Channel].ActualDestination - Data.Params[Channel].ActualPosition);
    }
  }
  return Value;
}

//  called from cIotModuleApp
void IRAM_ATTR cIotModuleSte::SetDestination(const uint8_t Channel, const uint32_t Destination)
{
  if (Channel != 0xff)
  {
    if (Data.Params[Channel % MAX_CHANNELS].Attached)
    {
      //  if reset inactive
      if (!Data.Params[Channel].ResetActive)
      {
        //  if reset request
        if (Destination == 0xffffffff)
        {
          //  set reset active flag and do not store position because this is an invalid value
          Data.Params[Channel].ResetActive = true;
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SetDestination - ResetRequest - ch : %d", __FILENAME__, __LINE__, Channel);
        }
        else
        {
          //  store actual destination
          uint32_t v = AddOffset(Destination);
          //  sign extend int32 to int64
          Data.Params[Channel].ActualDestination = v & 0x80000000 ? v | 0xffffffff00000000 : v;
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - SetDestination - ch : %d, v : %016llx", __FILENAME__, __LINE__, Channel, Data.Params[Channel].ActualDestination);
        }
      }
      //  if reset active
      else
      {
        //  if no request
        if (Destination != 0xffffffff)
        {
          //  store actual destination and set it as current position
          uint32_t v = AddOffset(Destination);
          //  sign extend int32 to int64
          Data.Params[Channel].ActualDestination = v & 0x80000000 ? v | 0xffffffff00000000 : v;
          Data.Params[Channel].ActualPosition = Data.Params[Channel].Extend_16_64.Set(Data.Params[Channel].ActualDestination);
          //  reset inactive
          Data.Params[Channel].ResetActive = false;
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SetDestination - ResetDone - ch : %d", __FILENAME__, __LINE__, Channel);
        }
      }
    }
  }
}

//  called from cIotModuleApp
void cIotModuleSte::SetControlPin(const uint8_t Channel, const uint8_t Type, const gpio_num_t Name, const gpio_pull_mode_t ResetMode)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SetControlPin - Type : %d, Name : %d", __FILENAME__, __LINE__, Type, Name);
  switch(Type)
  {
    case PIN_STEP:
      Data.Params[Channel % MAX_CHANNELS].PinStepName = Name;
      Data.Params[Channel % MAX_CHANNELS].PinStepResetMode = ResetMode;
      break;
    case PIN_DIR:
      Data.Params[Channel % MAX_CHANNELS].PinDirName = Name;
      Data.Params[Channel % MAX_CHANNELS].PinDirResetMode = ResetMode;
      break;
  }
}

//  called from cIotModuleApp
bool cIotModuleSte::Attach(uint8_t LedcTimer, uint8_t LedcChannel, uint8_t PcntChannel)
{
  if (!Data.Params[LedcTimer % MAX_CHANNELS].Attached)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - LedcTimer : %d, LedcChannel : %d, PcntChannel : %d", __FILENAME__, __LINE__, LedcTimer, LedcChannel, PcntChannel);

    //  *************** DO NOT MODIFY FOLLOWING FUNCTION CALL ORDER OTHERWISE THE BEHAVIOUR CHANGES BADLY ******************

    //  prepare and then apply the pcnt unit configuration
    pcnt_unit.pulse_gpio_num = Data.Params[LedcTimer].PinStepName,
    pcnt_unit.ctrl_gpio_num  = Data.Params[LedcTimer].PinDirName,
    pcnt_unit.unit           = (pcnt_unit_t)PcntChannel,
    pcnt_unit.channel        = PCNT_CHANNEL_0,
    pcnt_unit.pos_mode       = PCNT_COUNT_INC,
    pcnt_unit.neg_mode       = PCNT_COUNT_DIS,
    pcnt_unit.lctrl_mode     = PCNT_MODE_REVERSE,
    pcnt_unit.hctrl_mode     = PCNT_MODE_KEEP,
    pcnt_unit.counter_h_lim  = 0;
    pcnt_unit.counter_l_lim  = 0;
    ESP_ERROR_CHECK(pcnt_unit_config(&pcnt_unit));

    //  pause the counter
    ESP_ERROR_CHECK(pcnt_counter_pause((pcnt_unit_t)PcntChannel));

    //  prepare and then apply the ledc timer configuration
    ledc_timer.speed_mode      = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    ledc_timer.timer_num       = (ledc_timer_t)(LedcTimer % 4);
    ledc_timer.duty_resolution = LEDC_DUTY;
    ledc_timer.freq_hz         = START_SPEED;
    ESP_ERROR_CHECK(ledc_timer_config(&ledc_timer));

    //  immediately pause the counter timer
    ESP_ERROR_CHECK(ledc_timer_pause(LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(LedcTimer % 4)));

    //  prepare and then apply the ledc channel configuration, it sets implicitly gpio_num GPIO as output we use for step signal
    ledc_channel.speed_mode = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    //  odd channels not used
    ledc_channel.channel    = (ledc_channel_t)(LedcChannel % 8);
    ledc_channel.timer_sel  = (ledc_timer_t)(LedcTimer % 4);
    ledc_channel.intr_type  = LEDC_INTR_DISABLE;
    ledc_channel.gpio_num   = Data.Params[LedcTimer].PinStepName;
    ledc_channel.duty       = 1 << (LEDC_DUTY - 1);
    //  **************** set hpoint to 1 to get step GPIO low always after config *******************
    //  ********** if hpoint set to 0 step GPIO goes high after first use of ledc channel ***********
    ledc_channel.hpoint     = 1;
    ESP_ERROR_CHECK(ledc_channel_config(&ledc_channel));

    //  close step GPIO output to input to feed pcnt counter with pwm signal
    PIN_INPUT_ENABLE(GPIO_PIN_MUX_REG[Data.Params[LedcTimer].PinStepName]);

    //  clear and resume the counter
    ESP_ERROR_CHECK(pcnt_counter_clear((pcnt_unit_t)PcntChannel));
    ESP_ERROR_CHECK(pcnt_counter_resume((pcnt_unit_t)PcntChannel));

    //  acquire current position counter
    int16_t v;
    ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)PcntChannel, &v));
    Data.Params[LedcTimer].Extend_16_64.Reset(v);

    //  enable dir GPIO signal used also to control pcnt counter direction
    //  remove ??? ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[Channel].PinDirName));
    ESP_ERROR_CHECK(gpio_set_direction(Data.Params[LedcTimer].PinDirName, GPIO_MODE_INPUT_OUTPUT));
    ESP_ERROR_CHECK(gpio_set_level(Data.Params[LedcTimer].PinDirName, 0));

    //  init values
    Data.Params[LedcTimer].PreviousClockSignalEnabled = false;
    Data.Params[LedcTimer].PreviousSpeed = 0;
    Data.Params[LedcTimer].PreviousDirection = DIR_POS;
    Data.Params[LedcTimer].PreviousStatus = IDLE;
    Data.Params[LedcTimer].ResetActive = true;

    //  example of how to invert GPIO 25
    //  REG_SET_FIELD(GPIO_FUNC25_OUT_SEL_CFG_REG, GPIO_FUNC25_OUT_INV_SEL, 1);
    //  other examples
    //  REG_WRITE(LEDC_HSCH0_CONF0_REG, 0);
    //  REG_SET_FIELD(LEDC_HSCH0_HPOINT_REG, LEDC_HPOINT_HSCH0, 0);
    //  REG_SET_FIELD(LEDC_HSCH0_DUTY_REG, LEDC_DUTY_HSCH0, 0);

    Data.Params[LedcTimer].Attached = true;
  }
  return true;
}

bool cIotModuleSte::Detach(const uint8_t LedcTimer)
{
  if (LedcTimer != 0xff)
  {
    if (Data.Params[LedcTimer % MAX_CHANNELS].Attached)
    {
      //  odd channels not used
      ESP_ERROR_CHECK(ledc_stop(LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)((LedcTimer * 2) % 8), 0));
      
      //  open output mux to input mux? gpio_reset_pin will do it?
      PIN_INPUT_DISABLE(GPIO_PIN_MUX_REG[Data.Params[LedcTimer].PinStepName]);

      //  open output mux to input mux? gpio_reset_pin will do it?
      PIN_INPUT_DISABLE(GPIO_PIN_MUX_REG[Data.Params[LedcTimer].PinDirName]);
      
      ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[LedcTimer].PinStepName));
      
      ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[LedcTimer].PinDirName));

      Data.Params[LedcTimer].Attached = false;
    }
  }
  return true;
}

//  called from cIotModuleApp
void cIotModuleSte::Stop(const uint8_t Channel)
{
  if (Channel != 0xff)
  {
    if (Data.Params[Channel % MAX_CHANNELS].Attached)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Stop : ch : %d", __FILENAME__, __LINE__, Channel);
      Data.Params[Channel].ResetActive = true;
    }
  }
}

uint32_t IRAM_ATTR cIotModuleSte::AddOffset(const uint32_t Value)
{
  return Value + 0x80000000;
}

void IRAM_ATTR cIotModuleSte::SetCounterDirection(const uint8_t Channel, const DIR Direction)
{
  if (Direction == DIR_POS)
  {
    ESP_ERROR_CHECK(gpio_set_level(Data.Params[Channel].PinDirName, 1));
  }
  else
  {
    ESP_ERROR_CHECK(gpio_set_level(Data.Params[Channel].PinDirName, 0));
  }
}

void IRAM_ATTR cIotModuleSte::SetSpeed(const uint8_t Channel, const uint32_t Speed)
{
  uint32_t _Speed = Speed;
  if (_Speed < START_SPEED) _Speed = START_SPEED;
  if (_Speed > MAX_SPEED) _Speed = MAX_SPEED;
  ESP_ERROR_CHECK(ledc_set_freq(Channel / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(Channel % 4), _Speed));
}

// ====================================================================================
//  TODO change these global variables into cIotModuleSte members variables
uint8_t test = 0;
// ====================================================================================

void IRAM_ATTR cIotModuleSte::TimedInterpolate()
{
  // ====================================================================================
  //                             AVERAGE LOOP TIME
  // ====================================================================================
  int64_t uSec = esp_timer_get_time();
  AverageLoopTimeMS += ((uSec - OldUS) / 1000.0 - AverageLoopTimeMS) * 0.05;
  //  AverageLoopTimeMS = (uSec - OldUS) / 1000.0;
  OldUS = uSec;
  
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    if (Data.Params[Channel].Attached && !Data.Params[Channel].ResetActive)
    {
      // ====================================================================================
      //                                  STEP CONTROL
      // ====================================================================================
      //  Speed (steps per sec.)
      int64_t ActualSpeed = Data.Params[Channel].PreviousSpeed;
      //  Direction
      DIR ActualDirection = Data.Params[Channel].PreviousDirection;
      //  Status
      STATUS ActualStatus = Data.Params[Channel].PreviousStatus;
      //  ClockSignalEnabled
      bool ActualClockSignalEnabled = Data.Params[Channel].PreviousClockSignalEnabled;
      //  Max speed (steps per sec.)
      const uint32_t MaxSpeed = Data.Params[Channel].SpeedHz;
      //  Hz/mS (normal 1..3200) (max 65535)
      const uint16_t MaxAcc = Data.Params[Channel].AccHz_mSec * cIotModuleCom::TIMER_MS;
      //  Position (steps) reading the counter
      int16_t v;
      ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)Channel, &v));
      Data.Params[Channel].ActualPosition = Data.Params[Channel].Extend_16_64.Update(v);
      //  Destination (steps)
      const int64_t ActualDestination = Data.Params[Channel].ActualDestination;
      //  Distance signed (steps)
      int64_t Distance = ActualDestination - Data.Params[Channel].ActualPosition;
      //  Distance ABS
      int64_t DistanceAbs = Distance;
      if (Distance < 0) DistanceAbs = -Distance;
      //  braking velocity squared
      uint64_t BrakeSpeedSq = DistanceAbs * MaxAcc * 2;
      
      bool Accelerate = false;
   
      //  if actual speed squared < braking velocity squared
      if ((ActualSpeed * ActualSpeed * AverageLoopTimeMS * 1.05) / 1000 < BrakeSpeedSq)
      {
        //  if required positive movement
        if (Distance > 0)
        {
          //  if doing positive movement or idle
          if (ActualDirection == DIR_POS || ActualStatus == IDLE)
          {
            ActualDirection = DIR_POS;
            Accelerate = true;
          }
        }
        //  if required negative movement
        else if (Distance < 0)
        {
          //  if doing negative movement or idle
          if (ActualDirection == DIR_NEG || ActualStatus == IDLE)
          {
            ActualDirection = DIR_NEG;
            Accelerate = true;
          }
        }
      }

      //  if accelerate
      if (Accelerate)
      {
        //  accelaration with MaxAcc
        //ActualSpeed += MaxAcc;
        //  accelaration proportional to distance
        uint64_t acc = (MaxAcc * DistanceAbs) / 30;
        if (acc > MaxAcc) acc = MaxAcc;
        ActualSpeed += acc;
        //  limit to MaxSpeed
        if (ActualSpeed > MaxSpeed) ActualSpeed = MaxSpeed;
        //  start immediately with the minimum speed
        if (ActualSpeed < START_SPEED) ActualSpeed = START_SPEED;
      }
      //  if decelerate
      else
      {        
        if (ActualSpeed < START_SPEED)
        {
          //   stop
          ActualSpeed = 0;
        }
        else
        {
          ActualSpeed -= MaxAcc;
          if (ActualSpeed < START_SPEED) ActualSpeed = 0;
        }         
      }

      switch(ActualStatus)
      {
        case RUN_IDLE:
          if (gpio_get_level(Data.Params[Channel].PinStepName) == 0)
          {
            ESP_ERROR_CHECK(ledc_timer_pause(Channel / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(Channel % 4)));
            ActualStatus = IDLE;
          }
          break;
        case IDLE:
          if (Accelerate)
          {
            if (Distance > 0) SetCounterDirection(Channel, DIR_POS);
            else SetCounterDirection(Channel, DIR_NEG);
            ActualStatus = IDLE_RUN;
          }
          break;
        case IDLE_RUN:
          ESP_ERROR_CHECK(ledc_timer_resume(Channel / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(Channel % 4)));
          ActualStatus = RUN;
          break;
        case RUN:
          SetSpeed(Channel, ActualSpeed);
          if (ActualSpeed == 0) ActualStatus = RUN_IDLE;
          break;
      }

      // ----------------------------------------------- save values for the next time
      Data.Params[Channel].PreviousSpeed = ActualSpeed;
      Data.Params[Channel].PreviousDirection = ActualDirection;
      Data.Params[Channel].PreviousStatus = ActualStatus;
      Data.Params[Channel].PreviousClockSignalEnabled = ActualClockSignalEnabled;
      
#ifdef IOTMODULE_DEBUG      
      if (Channel == 0)
      {
        //  int32_t DeltaDistance = Data.Params[Channel].ActualPosition - Data.Params[Channel].PreviousPosition;
        //  int32_t DistanceLog = Distance;
        //  if (DistanceLog > 100) DistanceLog = 100;
        //  if (DistanceLog < -100) DistanceLog = -100;
        //  printf("%d %d %d\n", Data.Params[Channel].PreviousDirection, DistanceLog, ActualClockSignalEnabled);
        //  printf("%d %d %d %d\n", ActualDirection, Distance, ActualClockSignalEnabled, ActualSpeed);
        //  printf("%d %d\n", DeltaDistance, Distance);
        
        // printf("%d %d %ld\n", ActualDestination, Data.Params[Channel].ActualPosition, (long int)ActualSpeed);
        
        // if (Status == IDLE && ActualSpeed == 0)
        // {
        //   test++;
        //   if (test >= 8) test = 0;
        // } 
        pIotModuleApp->GenericWriteFloat(37, ActualDestination);
        pIotModuleApp->GenericWriteFloat(38, Data.Params[Channel].ActualPosition);
        pIotModuleApp->GenericWriteFloat(34, ActualSpeed);
        pIotModuleApp->GenericWriteFloat(35, AverageLoopTimeMS);
        pIotModuleApp->GenericWriteFloat(1, Status);
      }
#endif
    }
  }
  //  pIotModuleApp->GenericWriteFloat(1, micros() - OldUS);
}

void cIotModuleSte::PreFuota(void)
{
  //  Stop any interrupt activity
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    if (Data.Params[Channel].Attached)
    {
      //  clear attached flag to avoid interrupt entering TimedInterpolate for this channel
      Data.Params[Channel].Attached = false;
      //  detach channel
      Detach(Channel);
    }
  }
}
 
  